#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import pandas as pd

from sciveo.media.pipelines.processors.sci.base import *


class ProjectDatasetPlots(SciBaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.s3 = boto3.client('s3')
    self.default.update({
      "x_col": None,
      "time_column": None,
      "time_format": "%Y-%m-%d %H:%M:%S",
      "plots": [],
      "fft_plots": [],
      "plot_from": 0,
      "plot_to": -1
    })

  def process(self, content):
    try:
      if self.job['content_type'] == "project":
        project_path = os.path.join(self.base_tmp_path, "projects", self.job['content_id'])
        mkdirs(project_path)

        # Datasets with S3 located files
        if "bucket" in content and "key" in content:
          remote_path = f"{content['owner']}/{content['key']}"
          local_file_name = content['key'].replace("/", "-")
          local_path = os.path.join(project_path, local_file_name)

          if os.path.isfile(local_path):
            debug("SKIP", local_path)
          else:
            debug("Dataset S3 DWN", content["bucket"], remote_path, local_path)
            self.s3.download_file(content["bucket"], remote_path, local_path)

          file_name, file_extension = os.path.splitext(local_file_name)

          # Process CSV files
          if file_extension == ".csv":
            df = pd.read_csv(local_path)
            # Time Series
            debug(f"time_column[{self['time_column']}] time_format[{self['time_format']}]")
            if self["time_column"] is not None:
              df[self["time_column"]] = pd.to_datetime(df[self["time_column"]], format=self["time_format"])
              df = df.set_index(self["time_column"]).sort_index()

              # Plot FFT
              for y_col in self["fft_plots"]:
                try:
                  self.content_next_df_fft(content, df, y_col, project_path)
                except Exception as e:
                  exception(e)

            else:
              x_col = content['data'].get("dataset", {}).get("x_col", self["x_col"])
              df.set_index(x_col)

            list_plots =  content['data'].get("render", {}).get("plots", self["plots"])
            debug("plots", list_plots)
            for y_col in list_plots:
              try:
                self.content_next_df_plot(content, df[self["plot_from"]:self["plot_to"]], y_col, project_path, "time series")
              except Exception as e:
                exception(e)

            text = f"""
              <h3 class='sci-section'>Shape {df.shape}</h3>
              <h2 class='sci-section'>head/tail of dataset</h2>
              <div>{self.df_to_html(pd.concat([df.head(), df.tail()]))}</div>
              <h2 class='sci-section'>describe of dataset</h2>
              <div>{self.describe_df(df)}</div>
            """
            content.setdefault("next", [])
            text_content = self.content_text(content, "dataset plots and tables", text)
            content["next"].append(text_content)

          MediaJobState.queue().inc_progress(self.job["id"], 10)
    except Exception as e:
      exception(e)
    return content

  def content_type(self):
    return "dataset"

  def name(self):
    return "project-datasets-plots"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import subprocess
import time
import json
import datetime
import socket
import psutil
import platform
import re
import uuid
import numpy as np

from sciveo.tools.logger import *
from sciveo.tools.daemon import DaemonBase
from sciveo.tools.hardware import *
from sciveo.tools.formating import format_memory_size
from sciveo.api.base import APIRemoteClient


class MonitorTools:
  @staticmethod
  def serial():
    machine_serial = ""
    s = "-"
    list_uid_calls = [socket.gethostname, psutil.cpu_count, platform.processor]
    for uid_call in list_uid_calls:
      try:
        machine_serial += f"{uid_call()}{s}"
        s = ""
      except Exception:
        pass
    return re.sub(r'[\s,]', '', machine_serial)


class BaseMonitor(DaemonBase):
  def __init__(self, period=5, output_path=None):
    super().__init__(period=period)
    self.output_path = output_path
    if self.output_path is not None and os.path.isdir(self.output_path):
      self.output_path = os.path.join(self.output_path, "sciveo_monitor.json")

    self.data = {
      "CPU": {},
      "RAM": {},
      "DISK": {},
      "NET": {},
      "TEMP": {},
      "LOG": {},
      "INFO": {},
    }
    self.data["INFO"] = HardwareInfo()()
    self.data["INFO"].setdefault("CPU", {})
    self.list_logs = []

    self.api = APIRemoteClient()

    # Warmup the psutil
    psutil.cpu_percent(interval=0.3, percpu=True)
    initial_cpu_usage = psutil.cpu_percent(interval=None, percpu=True)
    self.previous_io_counters = {
      "DISK": psutil.disk_io_counters(perdisk=False),
      "NET": psutil.net_io_counters(pernic=False)
    }
    self.previous_time = {"DISK": time.time(), "NET": time.time()}
    time.sleep(1)

    self.data["serial"] = MonitorTools.serial()

    debug(f"init monitor with period={period}", self.data["serial"], "initial_cpu_usage", initial_cpu_usage)

  def __call__(self):
    return self.data

  def output_write(self):
    if os.path.isfile(self.output_path):
      with open(self.output_path, "rb+") as fp:
        fp.seek(-2, 2)
        fp.truncate()
      with open(self.output_path, "a") as fp:
        fp.write(f", {json.dumps(self.data)}]\n")
    else:
      with open(self.output_path, "w") as fp:
        fp.write(json.dumps([self.data]) + "\n")

  def loop(self):
    self.get_cpu_usage()
    self.get_temperatures()
    self.get_memory()
    self.get_gpu()
    self.get_disk()
    self.get_network()

    self.tail_logs()
    self.data["local_time"] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    api_result = self.api.POST_SCI("monitor", {"data": self.data})

    try:
      if self.output_path is not None:
        self.output_write()
    except Exception as e:
      error(e, "writing failed for", self.output_path)

    debug(self(), "api_result", api_result)

  def tail_logs(self):
    for log_name, log_path in self.list_logs:
      self.data["LOG"][log_name] = self.tail_file(log_path)[-3:]

  def get_cpu_usage(self):
    try:
      usage_per_core = psutil.cpu_percent(interval=None, percpu=True)
      self.data["CPU"]["usage per core"] = usage_per_core
      self.data["CPU"]["usage"] = float(np.array(usage_per_core).mean())
    except Exception:
      pass

  def get_temperatures(self):
    try:
      temperatures = psutil.sensors_temperatures()
      for k, v in temperatures.items():
        self.data["TEMP"][k] = [t.current for t in v]
    except Exception as e:
      pass

  def get_memory(self):
    try:
      memory = psutil.virtual_memory()
      self.data["RAM"]["used"] = memory.used
      self.data["RAM"]["total"] = memory.total
      self.data["RAM"]["free"] = memory.free
      # self.data["RAM"]["installed"] = format_memory_size(memory.total)
      self.data["INFO"]["RAM"] = f"total: {format_memory_size(memory.total)} used: {format_memory_size(memory.used)}"
    except Exception:
      pass

  def tail_file(self, file_path, block_size=1024):
    result = ["EMPTY"]
    try:
      with open(file_path,'rb') as fp:
        fp.seek(-block_size, os.SEEK_END)
        result = str(fp.read(block_size).rstrip()).split("\\n")
    except Exception as e:
      error(e, "tail_file", file_path)
    return result

  # Currently simple nvidia-smi wrapper impl
  def get_gpu(self):
    try:
      result = subprocess.run(
        [
          'nvidia-smi',
          '--query-gpu=gpu_uuid,gpu_name,index,power.draw,fan.speed,memory.total,memory.used,memory.free,utilization.gpu,utilization.memory,temperature.gpu',
          '--format=csv,nounits'
        ],
        capture_output=True, text=True, check=True
      )

      lines = result.stdout.strip().split('\n')
      self.data["GPU"] = {
        "raw_lines": lines[:7] # TODO: FIX this! Due to aws timestream constraints.
      }
    except Exception as e:
      pass

  def get_disk(self):
    try:
      list_metrics = ["read bytes", "write bytes", "read count", "write count", "read time", "write time"]
      disk_io_counters = psutil.disk_io_counters(perdisk=False)
      self.get_io_metrics("DISK", list_metrics, disk_io_counters)

      disk_usage = psutil.disk_usage('/')._asdict()
      self.data["INFO"]["DISK"] = f"{disk_usage['percent']}% ({round(disk_usage['used'] / (1024 * 1024 * 1024), 1)} GB / {round(disk_usage['total'] / (1024 * 1024 * 1024), 1)} GB)"
    except Exception as e:
      pass

  def get_network(self):
    try:
      list_metrics = ["bytes sent", "bytes recv", "packets sent", "packets recv"]
      net_io_counters = psutil.net_io_counters(pernic=False)
      self.get_io_metrics("NET", list_metrics, net_io_counters)
    except Exception as e:
      pass

  def get_io_metrics(self, name, list_metrics, io_counters):
    counters = io_counters._asdict()
    prev_io_counters = self.previous_io_counters[name]._asdict()
    time_diff = time.time() - self.previous_time[name]
    for metric_name in list_metrics:
      metric_key = metric_name.replace(' ', '_')
      self.data[name][metric_name] = counters[metric_key] - prev_io_counters[metric_key]

      if "bytes" in metric_name:
        speed_metric = metric_name.replace("bytes", "speed")
        self.data[name][speed_metric] = self.data[name][metric_name] / time_diff
    self.previous_time[name] = time.time()
    self.previous_io_counters[name] = io_counters

if __name__ == "__main__":
  debug(MonitorTools.serial())

  mon = BaseMonitor(period=10)
  mon.start()

  while(True):
    time.sleep(30)


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import math
import unittest

from sciveo.ml.nlp.tokenizers.bpe import *


class TestTokenizers(unittest.TestCase):
  def test_BPE(self):
    text = "節樂，《漢語大詞典》一則：「《史記．樂書》：凡作樂者，所以節槳。張守義正義：音洛，言不樂至荒淫也, 網站有中、英文版本，也有繁、簡體版，可通過每頁左上角的連結隨時調整 Ｕｎｉｃｏｄｅ! 🅤🅝🅘🅒🅞🅓🅔‽ 🇺‌🇳‌🇮‌🇨‌🇴‌🇩‌🇪! 😄 The very name strikes fear and awe into the hearts of programmers worldwide. We all know we ought to “support Unicode” in our software (whatever that means—like using wchar_t for all the strings, right?)"
    text += "Using a row in the above table to encode a code point less than 'First code point' (thus using more bytes than necessary) is termed an overlong encoding. These are a security problem because they allow the same code point to be encoded in multiple ways. Overlong encodings (of ../ for example) have been used to bypass security validations in high-profile products including Microsoft's IIS web server[14] and Apache's Tomcat servlet container.[15] Overlong encodings should therefore be considered an error and never decoded. Modified UTF-8 allows an overlong encoding of U+0000."

    T = BPETokenizer(max_size=512)
    T.train(text)

    t = "你好世界，美好的一天"
    self.assertTrue(T.decode(T.encode(t)) == t)

    t = "hello world and testing"
    self.assertTrue(T.decode(T.encode(t)) == t)


if __name__ == '__main__':
  unittest.main()


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import zipfile
import boto3

from sciveo.tools.logger import *
from sciveo.media.pipelines.processors.base import *


class FileArchiveZIP(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.s3 = boto3.client('s3')

  def upload(self, content_output):
    zip_remote_path = f"{content_output['owner']}/{content_output['key']}"
    debug("AWS S3 UPD", content_output["local_path"], "=>", content_output["bucket"], zip_remote_path)
    self.s3.upload_file(content_output["local_path"], content_output["bucket"], zip_remote_path)
    os.remove(content_output["local_path"])
    debug("RM", content_output["local_path"])

  def run(self, job, input):
    zip_file_name = f"{job['content_id']}.zip"
    zip_key = zip_file_name
    zip_path = os.path.join(self.base_tmp_path, zip_file_name)
    progress_per_media = self.max_progress / max(1, len(input))
    debug("run", job["id"], progress_per_media, "input", input)
    try:
      with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for i, media in enumerate(input):
          if self.is_content_type(media) and zip_key != media['key']:
            try:
              if "bucket" in media:
                remote_path = f"{media['owner']}/{media['key']}"
                file_name = remote_path.replace("/", "-")
                local_path = os.path.join(self.base_tmp_path, file_name)
                if os.path.isfile(local_path):
                  debug("SKIP", local_path)
                else:
                  debug("process_job content, AWS S3 DWN", media["bucket"], remote_path, local_path)
                  self.s3.download_file(media["bucket"], remote_path, local_path)
                zipf.write(local_path, arcname=file_name)
            except Exception as e:
              exception(e, media)
          else:
            debug("Media SKIP", media)
          MediaJobState.queue().inc_progress(job["id"], progress_per_media)

      content_output = {
        "name": f"ZIP {job['content_name']}",
        "guid": f"zip-{job['content_id']}",
        "content_type": "file",
        "local_path": zip_path,
        "key": zip_key,
        "bucket": "smiveo-file",
        "owner": job["owner"],
        "parent": job['content_id'],
        "processor": self.name(),
        "layout": {"name": self.name(), **self["layout"]},
        "data": {
          "info": {
            "size": os.path.getsize(zip_path)
          }
        }
      }

      self.upload(content_output)
      # job["output"] += content_output

      job["append-content"].append(content_output)

    except Exception as e:
      exception(e, zip_path)
    return []

  def content_type(self):
    return None

  def name(self):
    return "archive-zip"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import json
import threading


class GlobalConfiguration:
  config = None
  lock_config = threading.Lock()

  def __init__(self, name) -> None:
    home = os.path.expanduser('~')
    self.base_path = os.path.join(home, f'.{name}')
    self.data = {}

    self.default = {
      "api_base_url": "https://sciveo.com",
      "log_min_level": "DEBUG"
    }

    try:
      self.read_local_files()
      self.read_environment()
    except Exception as e:
      error("Exception", e)

  @staticmethod
  def get(name='sciveo', reload=False):
    with GlobalConfiguration.lock_config:
      if GlobalConfiguration.config is None or reload:
        GlobalConfiguration.config = GlobalConfiguration(name)
      return GlobalConfiguration.config

  @staticmethod
  def set(k, v):
    with GlobalConfiguration.lock_config:
      GlobalConfiguration.config.data[k] = v

  def __getitem__(self, key):
    key = key.lower()
    return self.data.get(key, self.default.get(key, None))

  def save_kv(self, key, value):
    key = key.strip().lower().replace("sciveo_", "")
    value = value.strip()
    self.data[key] = value

  def read_environment(self):
    for k, v in os.environ.items():
      self.save_kv(k, v)

  def read_file(self, path):
    with open(path, 'r') as fp:
      lines = fp.readlines()
      for line in lines:
        parts = line.strip().split('=')
        if len(parts) == 2:
          self.save_kv(parts[0], parts[1])

  def read_json(self, path):
    data_json = {}
    with open(path, 'r') as fp:
      data_json = json.load(fp)
    for key, value in data_json.items():
      self.save_kv(key, value)

  def read_local_files(self):
    if os.path.exists(self.base_path):
      for path, _, files in os.walk(self.base_path):
        for file_name in files:
          file_path = os.path.join(path, file_name)
          _, file_extension = os.path.splitext(file_name)
          if file_extension == ".json":
            self.read_json(file_path)
          else:
            self.read_file(file_path)


class ConfigurationArguments:
  def __init__(self, default, **kwargs) -> None:
    self.default = default
    self.arguments = {}
    for k, v in self.default.items():
      self.arguments[k] = kwargs.get(k, v)

  def __getitem__(self, key):
    return self.arguments[key]

  def __len__(self):
    return len(self.arguments)

  def __repr__(self):
    return self.arguments


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import cv2

from sciveo.tools.common import *
from sciveo.media.pipelines.processors.base import *


class VideoResize(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.current_heights = {}
    self.default.update({
      "CRF": 20, "heights": [720, 480],
      "downsampling_rate": 1,
    })

  def init_low_videos(self, frame):
    self.current_heights = {}
    for h in self["heights"]:
      h, w = get_frame_resolution(frame, h)
      if h in self.current_heights:
        continue
      self.current_heights[h] = [w, h]

      tag = f"resized-{h}"
      resized_local_path = self.add_suffix_to_filename(self.local_path, tag)
      self.low_resolutions.append((w, h))
      self.low_videos.append(VideoWriterFFMPEG.new(resized_local_path, w=w, crf=str(self["CRF"])))
      self.next_content(self.media, tag, resized_local_path, w=w, h=h)

  def process(self, media):
    self.media = media
    self.local_path = media["local_path"]

    self.low_videos = []
    self.low_resolutions = []

    self.cap = cv2.VideoCapture(self.local_path)
    n_frame = 0
    while(self.cap.isOpened()):
      n_frame += 1
      ret, frame = self.cap.read()
      if not ret:
        break

      if n_frame == 1:
        self.init_low_videos(frame)

      for i, low_video in enumerate(self.low_videos):
        if n_frame % self["downsampling_rate"] == 0:
          frame_resized = cv2.resize(frame, self.low_resolutions[i], interpolation=cv2.INTER_AREA)
          low_video.write(frame_resized)

    self.cap.release()
    for low_video in self.low_videos:
      low_video.close()

    return self.media

  def is_resizer(self):
    return True

  def content_type(self):
    return "video"

  def name(self):
    return "video-resize"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import socket
import threading
from scapy.all import sniff, IP, TCP

from sciveo.tools.logger import *
from sciveo.tools.timers import Timer


class StreamSniffer:
  def __init__(self, iface=None):
    self.iface = iface
    self.running = False
    self.lock = threading.Lock()
    self.streams = {}

  def start(self):
    self.running = True
    self.sniff_thread = threading.Thread(target=self.sniff_packets)
    self.sniff_thread.start()

  def stop(self):
    debug("stopping...")
    self.running = False
    self.sniff_thread.join()

  def sniff_packets(self):
    debug("start sniffing on", self.iface)
    sniff(iface=self.iface, prn=self.on_packet, stop_filter=self.should_stop)

  def on_packet(self, packet):
    if IP in packet:
      self.append_ip_packet(packet)

  def should_stop(self, packet):
    return not self.running

  def append_ip_packet(self, packet):
    ip_src = packet[IP].src
    with self.lock:
      self.streams.setdefault(ip_src, [])
      self.streams[ip_src].append(packet)

  def get_ip_stream(self, ip):
    current_packets = []
    with self.lock:
      if ip in self.streams:
        current_packets = self.streams[ip][:]
        self.streams[ip] = []
    return current_packets

  def keys(self):
    with self.lock:
      return list(self.streams.keys())


if __name__ == "__main__":
  # debug(NetworkTools(timeout=1.0, localhost=False).scan_port(port=9901))

  import time
  sniffer = StreamSniffer(iface="en0")
  sniffer.start()
  time.sleep(5)
  sniffer.stop()
  debug(sniffer.keys())


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import socket
import ipaddress
import threading
import json

from sciveo.tools.logger import *
from sciveo.tools.timers import Timer


class NetworkTools:
  def __init__(self, **kwargs):
    self.default_arguments = {
      "timeout": 1.0,
      "localhost": False,
      "host": None,
      "ports": []
    }

    self.arguments = {}
    for k, v in self.default_arguments.items():
      self.arguments[k] = kwargs.get(k, v)

    self.net_classes = ["192.168.", "10."]
    for i in range(16, 32):
      self.net_classes.append(f"172.{i}.")

    self.data = {}
    self.data_lock = threading.Lock()

    # self.known_ports = [20, 21, 22, 23, 25, 53, 67, 68, 69, 80, 110, 123, 143, 161, 162, 179, 194, 389, 443, 465, 514, 587, 636, 993, 995, 1080, 1194, 1433, 1521, 1723, 2049, 2082, 2083, 2086, 2087, 2095, 2096, 2181, 2375, 2376, 3000, 3306, 3389, 5432, 5900, 5984, 6379, 8080, 8443, 8888, 9000, 9090, 9200, 9300, 11211, 27017]
    self.known_ports = {
      20: socket.SOCK_STREAM,  # FTP Data (TCP)
      21: socket.SOCK_STREAM,  # FTP Control (TCP)
      22: socket.SOCK_STREAM,  # SSH (TCP)
      23: socket.SOCK_STREAM,  # Telnet (TCP)
      25: socket.SOCK_STREAM,  # SMTP (TCP)
      53: socket.SOCK_DGRAM,   # DNS (UDP, but also uses TCP for large queries)
      67: socket.SOCK_DGRAM,   # DHCP Server (UDP)
      68: socket.SOCK_DGRAM,   # DHCP Client (UDP)
      69: socket.SOCK_DGRAM,   # TFTP (UDP)
      80: socket.SOCK_STREAM,  # HTTP (TCP)
      110: socket.SOCK_STREAM, # POP3 (TCP)
      123: socket.SOCK_DGRAM,  # NTP (UDP)
      143: socket.SOCK_STREAM, # IMAP (TCP)
      161: socket.SOCK_DGRAM,  # SNMP (UDP)
      162: socket.SOCK_DGRAM,  # SNMP Trap (UDP)
      179: socket.SOCK_STREAM, # BGP (TCP)
      194: socket.SOCK_STREAM, # IRC (TCP)
      389: socket.SOCK_STREAM, # LDAP (TCP)
      443: socket.SOCK_STREAM, # HTTPS (TCP)
      465: socket.SOCK_STREAM, # SMTPS (TCP)
      514: socket.SOCK_DGRAM,  # Syslog (UDP)
      587: socket.SOCK_STREAM, # SMTP Submission (TCP)
      636: socket.SOCK_STREAM, # LDAPS (TCP)
      993: socket.SOCK_STREAM, # IMAPS (TCP)
      995: socket.SOCK_STREAM, # POP3S (TCP)
      1080: socket.SOCK_STREAM, # SOCKS (TCP)
      1194: socket.SOCK_DGRAM,  # OpenVPN (UDP)
      1433: socket.SOCK_STREAM, # MS-SQL (TCP)
      1521: socket.SOCK_STREAM, # Oracle DB (TCP)
      1723: socket.SOCK_STREAM, # PPTP (TCP)
      2049: socket.SOCK_DGRAM,  # NFS (UDP)
      2082: socket.SOCK_STREAM, # cPanel (TCP)
      2083: socket.SOCK_STREAM, # cPanel SSL (TCP)
      2086: socket.SOCK_STREAM, # WHM (TCP)
      2087: socket.SOCK_STREAM, # WHM SSL (TCP)
      2095: socket.SOCK_STREAM, # Webmail (TCP)
      2096: socket.SOCK_STREAM, # Webmail SSL (TCP)
      2181: socket.SOCK_STREAM, # ZooKeeper (TCP)
      2375: socket.SOCK_STREAM, # Docker (TCP)
      2376: socket.SOCK_STREAM, # Docker SSL (TCP)
      3000: socket.SOCK_STREAM, # Node.js (TCP)
      3306: socket.SOCK_STREAM, # MySQL (TCP)
      3389: socket.SOCK_STREAM, # RDP (TCP)
      5432: socket.SOCK_STREAM, # PostgreSQL (TCP)
      5900: socket.SOCK_STREAM, # VNC (TCP)
      5984: socket.SOCK_STREAM, # CouchDB (TCP)
      6379: socket.SOCK_STREAM, # Redis (TCP)
      8080: socket.SOCK_STREAM, # HTTP Alt (TCP)
      8443: socket.SOCK_STREAM, # HTTPS Alt (TCP)
      8888: socket.SOCK_STREAM, # cPanel Alt (TCP)
      9000: socket.SOCK_STREAM, # PHP-FPM (TCP)
      9090: socket.SOCK_STREAM, # Prometheus (TCP)
      9200: socket.SOCK_STREAM, # Elasticsearch (TCP)
      9300: socket.SOCK_STREAM, # Elasticsearch Cluster (TCP)
      11211: socket.SOCK_DGRAM, # Memcached (UDP)
      27017: socket.SOCK_STREAM, # MongoDB (TCP)
    }

  def get_local_nets(self):
    list_local_ips = []
    try:
      import netifaces
      interfaces = netifaces.interfaces()
      for interface in interfaces:
        addrs = netifaces.ifaddresses(interface)
        if netifaces.AF_INET in addrs:
          ip = addrs[netifaces.AF_INET][0]['addr']
          for net_class in self.net_classes:
            if ip.startswith(net_class):
              list_local_ips.append(ip)
    except Exception as e:
      warning("netifaces not installed")
    return list_local_ips

  def generate_ip_list(self, base_ip):
    octets = base_ip.split('.')
    network_prefix = '.'.join(octets[:3])
    return [f'{network_prefix}.{i}' for i in range(1, 255)]

  def scan_port(self, port=22, network=None):
    t = Timer()
    self.data["scan"] = {}
    self.data["scan"].setdefault(port, [])

    if network is None:
      list_local_ips = self.get_local_nets()
      # debug("scan_port", "list_local_ips", list_local_ips)
      for local_ip in list_local_ips:
        list_ip = self.generate_ip_list(local_ip)
        self.scan_port_hosts(list_ip, port)
    else:
      list_ip = []
      net = ipaddress.ip_network(network, strict=False)
      for ip in net.hosts():
        list_ip.append(str(ip))
      self.scan_port_hosts(list_ip, port)

    if self.arguments["localhost"]:
      self.scan_port_hosts(["127.0.0.1"], port)
    self.data["scan"][port].sort(key=lambda ip: int(ip.split('.')[-1]))
    info(f"scan_port [{port}] elapsed time {t.stop():.1f}s", self.data["scan"][port])
    return self.data["scan"][port]

  def scan_port_hosts(self, list_ip, port=22):
    timeout = self.arguments["timeout"]
    list_threads = []
    for ip in list_ip:
      t = threading.Thread(target=self.scan_host_port, args=(ip, port, timeout))
      t.start()
      list_threads.append(t)
    for t in list_threads:
      t.join()

  def test_socket(self, ip, port, timeout, transport=socket.SOCK_STREAM):
    try:
      with socket.socket(socket.AF_INET, transport) as sock:
        sock.settimeout(timeout)
        if transport == socket.SOCK_STREAM:
          result = sock.connect_ex((ip, port))
          if result == 0:
            return True
        elif transport == socket.SOCK_DGRAM:
          sock.sendto(b"", (ip, port))
          try:
            data, addr = sock.recvfrom(1)
            return True
          except socket.timeout:
            # warning("sock timeout", (ip, port, transport))
            pass
    except socket.error:
      pass
    return False

  def scan_host_port(self, ip, port, timeout):
    transport = self.known_ports.get(port, socket.SOCK_STREAM)
    if self.test_socket(ip, port, timeout, transport):
      with self.data_lock:
        if "scan" in self.data:
          self.data["scan"][port].append(ip)
        if "host" in self.data:
          self.data["host"][ip].append(port)

  def scan_host(self, host):
    timer = Timer()
    self.data.setdefault("host", {})
    self.data["host"][host] = []
    timeout = self.arguments["timeout"]
    list_threads = []
    if len(self.arguments["ports"]) == 0:
      list_ports = list(self.known_ports.keys())
    else:
      list_ports = self.arguments["ports"]
    for port in list_ports:
      thr = threading.Thread(target=self.scan_host_port, args=(host, port, timeout))
      thr.start()
      list_threads.append(thr)
    for thr in list_threads:
      thr.join()

    self.data["host"][host].sort(key=lambda port: int(port))
    info(f"scan host [{host}] elapsed time {timer.stop():.1f}s", self.data["host"][host])
    return self.data["host"][host]


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import os
import time
from flask import Flask, jsonify, render_template, request, url_for, Response, abort
from waitress import serve

from sciveo.tools.logger import *
from sciveo.tools.daemon import DaemonBase
from sciveo.tools.timers import Timer
from sciveo.tools.compress import CompressJsonData
from sciveo.tools.configuration import GlobalConfiguration
from sciveo.api.predictors import Predictors


CONFIG = GlobalConfiguration.get()
SCI_API_AUTH_TOKEN = CONFIG["api_auth_token"]
SCI_API_PREFIX = CONFIG["api_prefix"]


app = Flask(__name__)


def authenticate():
  token = request.headers.get('Authorization')
  if token != f"Bearer {SCI_API_AUTH_TOKEN}":
    abort(401, 'Unauthorized')

@app.before_request
def before_request_func():
  authenticate()

@app.route(f'/{SCI_API_PREFIX}/predict', methods=['POST'])
def predict_endpoint():
  result = {"error": "invalid data"}
  try:
    if request.is_json:
      data = request.json
      if "predictor" in data and "X" in data:
        t = Timer()
        predicted = Predictors.get().predict(data["predictor"], data["X"])

        if data.get("compressed", 0) > 0:
          predicted = CompressJsonData().compress(predicted)

        result = {
          data["predictor"]: predicted,
          "stats": {
            "elapsed": t.stop(),
            "count": len(data["X"])
          }
        }
  except Exception as e:
    exception("predict_endpoint", e)
  return jsonify(result)


class WebServerDaemon(DaemonBase):
  def __init__(self, host="0.0.0.0", port=8901, predictors=None):
    super().__init__()
    self.host = host
    self.port = port

    if predictors is not None:
      Predictors.set(predictors)

  def run(self):
    info("run", [self.host, self.port])
    Predictors.get().print()

    serve(app, host=self.host, port=self.port)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import matplotlib.pyplot as plt
import matplotlib.dates as mdates

import numpy as np
import pandas as pd

from sciveo.tools.logger import *
from sciveo.tools.array import *
from sciveo.media.ml.encoders.normalizer import *


class BaseDataSet:
  def __init__(self, data, columns=None):
    self.data = data
    self.normalizer = Normalizer()

    if columns is not None:
      self.data = self.data[columns]
    self.columns = self.data.columns

  def normalize(self):
    self.data = self.normalizer.transform(self.data)
  def denormalize(self, key):
    self.dataset[key] = self.normalizer.inverse(self.dataset[key])

  # Split dataset
  def split(self, ratios=[["train", 0.85], ["val", 0.10], ["test", 0.05]]):
    self.ratios = ratios
    self.dataset = {}
    prev_idx = 0
    for k, v in ratios:
      next_idx = prev_idx + int(self.data.shape[0] * v)
      self.dataset[k] = self.data[prev_idx:next_idx]
      prev_idx = next_idx

  def summary(self):
    return self.data.describe().transpose()[["count", "min", "max", "mean", "std", "25%", "50%", "75%"]]



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import boto3
import cv2
import io
import base64
from PIL import Image

import torch
from torchvision import models, transforms

import numpy as np
import pandas as pd

from transformers import AutoTokenizer, AutoModel

from sciveo.tools.logger import *
from sciveo.ml.base import BaseML


class TextEmbedding(BaseML):
  def __init__(self, model_name='softel/sentence-base-v0.3', cache_dir=None, device=None) -> None:
    super().__init__(model_name=model_name, cache_dir=cache_dir, device=device)
    self.pipe = None
    self.tokenizer = None
    self.max_tokens = 512
    self.overlap = 128
    self.hidden = False

  def load_model(self):
    debug("loading model", self.model_name)
    self.pipe = AutoModel.from_pretrained(
        self.model_name,
        cache_dir=self.cache_dir
    )#.to(self.device)
    self.tokenizer = AutoTokenizer.from_pretrained(
      self.model_name,
      cache_dir=self.cache_dir
    )

  def init(self):
    if self.pipe is None or self.tokenizer is None:
      self.load_model()
      self.post_init()

  def get_tokens_count(self, text):
    return len(self.tokenizer(text, padding=False, truncation=False, return_tensors=None)['input_ids'])

  def chunk_text(self, text):
    tokens = self.tokenizer(text, padding=False, truncation=False, return_tensors=None)['input_ids']
    chunks = []
    for i in range(0, len(tokens), self.max_tokens - self.overlap):
      chunk = tokens[i:i + self.max_tokens]
      chunks.append(self.tokenizer.decode(chunk, skip_special_tokens=True))
      if len(chunk) < self.max_tokens:
        break
    return chunks

  def get_chuncked_embeddings(self, text):
    chunks = self.chunk_text(text)
    embeddings = []
    for chunk in chunks:
      embedding = np.array(self.get_sentence_embedding(chunk))
      embeddings.append(embedding)
    return np.mean(np.array(embeddings), axis=0)

  def get_sentence_embedding(self, text):
    while(True):
      try:
        return self.embed(text)
      except:
        text = text[:len(text) - 8]

  def get_embedding(self, text):
    total_tokens = self.get_tokens_count(text)
    if total_tokens<=self.max_tokens:
      return self.get_sentence_embedding(text).tolist()
    else:
      return self.get_chuncked_embeddings(text).tolist()

  def get_chunks_embedding(self, chunks):
    embeddings = []
    for text in chunks:
      embeddings.append(self.get_embedding(text))
    embeddings = np.array(embeddings)
    return np.mean(embeddings, axis=0).tolist()

  def embed(self, text):
    inputs = self.tokenizer.encode_plus(text, return_tensors='pt')#.to(self.device)

    if self.hidden:
      with torch.no_grad():
        last_hidden_state = self.pipe(**inputs, output_hidden_states=True).hidden_states[-1]
        weights_for_non_padding = inputs['attention_mask'] * torch.arange(start=1, end=last_hidden_state.shape[1] + 1).unsqueeze(0)
        sum_embeddings = torch.sum(last_hidden_state * weights_for_non_padding.unsqueeze(-1), dim=1)
        num_of_none_padding_tokens = torch.sum(weights_for_non_padding, dim=-1).unsqueeze(-1)
        embeddings = sum_embeddings / num_of_none_padding_tokens
    else:
      outputs = self.pipe(**inputs)
      sequence_output = outputs[0]
      input_mask_expanded = inputs['attention_mask'].unsqueeze(-1).expand(sequence_output.size()).float()
      embeddings = torch.sum(sequence_output * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)
    return embeddings.detach().numpy()[0]

  def predict_one(self, x):
    self.init()
    if isinstance(x, list):
      return self.get_chunks_embedding(x)
    else:
      return self.get_embedding(x)

  def predict(self, X):
    self.init()
    predictions = []
    for current_x in X:
      embedding = self.predict_one(current_x)
      if not isinstance(embedding, list):
        embedding = embedding.tolist()
      predictions.append(embedding)
    return predictions


class SentenceEmbedding(TextEmbedding):
  def __init__(self, model_name='softel/sentence-bge-v1.3', cache_dir=None, device=None) -> None:
    super().__init__(model_name=model_name, cache_dir=cache_dir, device=device)
    self.model = None
    self.max_tokens = 8192
    self.normalize_embeddings = True

  def init(self):
    if self.model is None:
      from sentence_transformers import SentenceTransformer
      self.model = SentenceTransformer(self.model_name).to(self.device)
      self.post_init()

  def predict_one(self, x):
    return self.predict([x])[0]

  def predict(self, X):
    self.init()
    predictions = self.model.encode(X, normalize_embeddings=self.normalize_embeddings)
    return predictions.tolist()